'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.publishNode = exports.createNode = exports.createInitialNode = exports.createFinalNodeFromProvisionalNode = exports.createChildAccessor = exports.createNodeAccessor = undefined;

var _rx = require('rx');

var _rx2 = _interopRequireDefault(_rx);

var _hooks = require('./hooks');

var _tree = require('./tree');

var _utils = require('../utils');

var _errors = require('../errors');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var exposedNodeProps = ['reduce', 'asObservable', 'setInitialState', 'connect'];

var createNodeAccessor = exports.createNodeAccessor = function createNodeAccessor(node) {
  return function (path, value) {
    var keys = path.split('.');
    return publishNode(keys.reduce(function (acc, key, i) {
      if (typeof value !== 'undefined' && i === keys.length - 1) {
        return acc.child(key, value);
      }
      return acc.child(key);
    }, node));
  };
};

var createChildAccessor = exports.createChildAccessor = function createChildAccessor(addChildrenSubject, getChildrenSubject) {
  return function (key, value) {
    if (!getChildrenSubject) (0, _errors.throwChildrenError)();
    var beforeChildren = getChildrenSubject.getValue();
    var nodeSubject = beforeChildren[key];
    if (!nodeSubject || nodeSubject.getValue().provisional) {
      if (typeof value !== 'undefined') {
        addChildrenSubject.onNext({ key: key, value: value, provisional: false });
      } else if (!nodeSubject) {
        addChildrenSubject.onNext({ key: key, value: null, provisional: true });
      }
    }
    return getChildrenSubject.getValue()[key].getValue();
  };
};

var createFinalNodeFromProvisionalNode = exports.createFinalNodeFromProvisionalNode = function createFinalNodeFromProvisionalNode(_ref) {
  var observable = _ref.observable;
  var provisionalNode = _ref.provisionalNode;
  var setNextState = _ref.setNextState;
  var observableSubject = provisionalNode.observableSubject;
  var nodeSubject = provisionalNode.nodeSubject;


  if (observable) {
    observableSubject.onNext(observable);
  }

  var nodeProps = Object.assign({}, provisionalNode, {
    provisional: false
  });

  if (typeof setNextState === 'function') {
    nodeProps.setNextState = setNextState;
  }

  var node = Object.assign(createNodeAccessor(nodeProps), nodeProps);

  nodeSubject.onNext(node);

  return node;
};

var createInitialNode = exports.createInitialNode = function createInitialNode(_ref2) {
  var addChildrenSubject = _ref2.addChildrenSubject;
  var getChildrenSubject = _ref2.getChildrenSubject;
  var pauser = _ref2.pauser;
  var observable = _ref2.observable;
  var hookMap = _ref2.hookMap;
  var setNextState = _ref2.setNextState;
  var provisional = _ref2.provisional;

  var observableSubject = new _rx2.default.ReplaySubject(1);
  // Great name right?
  var combinedObservable = observableSubject.flatMapLatest(function (obs) {
    return obs;
  });
  var externalObservable = combinedObservable.replay();
  var child = createChildAccessor(addChildrenSubject, getChildrenSubject);
  var nodeSubject = new _rx2.default.BehaviorSubject();
  var asObservable = function asObservable() {
    return externalObservable;
  };
  var reduce = (0, _hooks.createReduce)(combinedObservable, nodeSubject.asObservable(), hookMap, pauser);
  var connect = function connect() {
    var allDisposable = new _rx2.default.CompositeDisposable();
    var children = getChildrenSubject && getChildrenSubject.getValue();
    if (children) {
      var keys = Object.keys(children);
      keys.forEach(function (key) {
        return allDisposable.add(children[key].getValue().connect());
      });
    }
    allDisposable.add(asObservable().connect());
    return allDisposable;
  };

  var setInitialState = function setInitialState(initialState) {
    var currentNode = nodeSubject.getValue();
    if (!currentNode.provisional) {
      (0, _errors.throwFinalizedError)();
    } else {
      (0, _tree.createTree)({
        initialState: initialState,
        createNode: createNode,
        pauser: pauser,
        hookMap: hookMap,
        provisional: false,
        provisionalNode: node
      });
      return publishNode(nodeSubject.getValue());
    }
  };

  if (observable) {
    observableSubject.onNext(observable);
  }

  var nodeProps = {
    addChildrenSubject: addChildrenSubject,
    getChildrenSubject: getChildrenSubject,
    reduce: reduce,
    child: child,
    setNextState: setNextState,
    nodeSubject: nodeSubject,
    provisional: !!provisional,
    provisionalNode: !!provisional && node,
    pauser: pauser,
    combinedObservable: combinedObservable,
    observableSubject: observableSubject,
    asObservable: asObservable,
    connect: connect,
    setInitialState: setInitialState
  };

  var node = Object.assign(createNodeAccessor(nodeProps), nodeProps);

  nodeSubject.onNext(node);

  return node;
};

var createNode = exports.createNode = function createNode(_ref3) {
  var addChildrenSubject = _ref3.addChildrenSubject;
  var getChildrenSubject = _ref3.getChildrenSubject;
  var pauser = _ref3.pauser;
  var observable = _ref3.observable;
  var hookMap = _ref3.hookMap;
  var setNextState = _ref3.setNextState;
  var provisional = _ref3.provisional;
  var provisionalNode = _ref3.provisionalNode;

  if (provisionalNode) {
    return createFinalNodeFromProvisionalNode({
      observable: observable,
      setNextState: setNextState,
      provisionalNode: provisionalNode
    });
  } else {
    return createInitialNode({
      addChildrenSubject: addChildrenSubject,
      getChildrenSubject: getChildrenSubject,
      pauser: pauser,
      observable: observable,
      hookMap: hookMap,
      setNextState: setNextState,
      provisional: provisional
    });
  }
};

var publishNode = exports.publishNode = function publishNode(node) {
  return Object.assign(function () {
    return node.apply(null, arguments);
  }, (0, _utils.pick)(node, exposedNodeProps));
};